//==============================================================================
//
//	s1d13c00fb.c
//
//	This file provides the setup routines for the Epson S1D13C00 frame buffer
//	driver.
//
//	WARNING: This file uses a chip-specific include file. The include file 
//	contains all hardware information needed to generate proper timings and 
//	access registers. For timing and register details, see the S1D13C00
//	Hardware Functional Specification.
//
//	©SEIKO EPSON CORPORATION 2024. All rights reserved.
//
//	This driver software is distributed as is, without any warranty of any
//	kind, either express or implied as further specified in the GNU Public
//	License. This software may be used and distributed according to the terms
//	of the GNU Public License, version 2. See the file COPYING in the main
//	directory of this archive for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.
//
//==============================================================================




#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/uaccess.h>
#include <linux/signal.h>
#include <linux/platform_device.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/timer.h>
#include "s1d13c00fb.h"




char* s1d13c00fb_version = "EPSON13C00FB: Revision: 1.00";

struct fb_rect
{
	__u32 dx;	// screen-relative
	__u32 dy;
	__u32 width;
	__u32 height;
};




// Local functions.
static int  s1d13c00fb_probe( struct platform_device* pd );
static int  s1d13c00fb_remove( struct platform_device* dev );
static int  s1d13c00fb_release( struct fb_info* fbp, int user );
static int  s1d13c00fb_checkVar( struct fb_var_screeninfo* var, struct fb_info* info );
static void s1d13c00fb_fillRect( struct fb_info* fbp, const struct fb_fillrect* fbr );
static void s1d13c00fb_copyArea( struct fb_info* fbp, const struct fb_copyarea* fbc );
static void s1d13c00fb_imageBlit( struct fb_info* fbp, const struct fb_image* fbi );
static void s1d13c00fb_deferredio( struct fb_info* info, struct list_head* pageList );
static void s1d13c00fb_deferredioTouch( struct fb_info* info, const struct fb_rect* rect );
static void s1d13c00fb_update_pages( struct fb_info* info, unsigned long offset, int nPages );
static void s1d13c00fb_flush_dirty_rect( struct fb_info* info );




static struct platform_driver s1d13c00fb_driver =
{
	.probe	= s1d13c00fb_probe,
	.remove = s1d13c00fb_remove,
	.driver	= {
				.owner	= THIS_MODULE,
				.name	= "s1d13c00fb",
			  },
};

static struct platform_device* s1d13c00fb_device;

static struct fb_ops s1d13c00fb_ops =
{
	.owner			= THIS_MODULE,
	.fb_release		= s1d13c00fb_release,
	.fb_check_var	= s1d13c00fb_checkVar,
	.fb_fillrect	= s1d13c00fb_fillRect,
	.fb_copyarea	= s1d13c00fb_copyArea,
	.fb_imageblit	= s1d13c00fb_imageBlit,
	.fb_mmap		= fb_deferred_io_mmap,
	#ifdef CONFIG_S1D13C00_IOCTL
	.fb_ioctl		= s1d13c00fb_ioctl,
	#endif
};

struct fb_fix_screeninfo s1d13c00fb_fix =
{
	.id 		 = "epson13c00fb", 
	.type		 = FB_TYPE_PACKED_PIXELS,
	.visual		 = FB_VISUAL_TRUECOLOR,
	.type_aux	 = 0,
	.xpanstep	 = 0,
	.ypanstep	 = 0,
	.ywrapstep	 = 0,
	.smem_len	 = VFB_SIZE_BYTES,
	.line_length = C00_DRIVER_STRIDE,
	.accel		 = FB_ACCEL_NONE,
};

static struct fb_deferred_io s1d13c00fb_defio =
{
	.delay			  = VIDEO_REFRESH_PERIOD,
	.sort_pagereflist = true,
	.deferred_io	  = s1d13c00fb_deferredio,
};




int s1d13c00fb_init( void )
//
//	Module initialization.
//
{
	int ret;

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	ret = platform_driver_register( &s1d13c00fb_driver );

	if ( !ret )
	{
		s1d13c00fb_device = platform_device_register_simple( "s1d13c00fb", 0, NULL, 0 );
		if ( IS_ERR(s1d13c00fb_device) )
		{
			printk( KERN_WARNING "%s(): ERROR! s1d13c00fb_device is not a device.\n", __FUNCTION__ );
			platform_driver_unregister( &s1d13c00fb_driver );
			ret = PTR_ERR( s1d13c00fb_device );
		}
	}
	return ret;
}




void s1d13c00fb_exit( void )
//
//	Module termination.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	platform_device_unregister( s1d13c00fb_device );
	platform_driver_unregister( &s1d13c00fb_driver );
}




static int s1d13c00fb_probe( struct platform_device* pd )
//
//	Entry point for frame buffer driver.
//
{
	struct fb_info*        info;
	struct s1d13c00fb_par* par;
	int                    retVal = 0;

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	info = framebuffer_alloc( sizeof(struct s1d13c00fb_par), &pd->dev );
	if ( !info )
		return -ENOMEM;

	par = info->par;

	// Allocate the virtual framebuffer
	info->screen_base = (char*)devm_kzalloc( &pd->dev, VFB_SIZE_BYTES, GFP_KERNEL );
	if ( info->screen_base == NULL )
	{
		framebuffer_release( info );
		printk( KERN_WARNING "%s(): devm_kzalloc() failed.\n", __FUNCTION__ );
		return -ENOMEM;
	}

	info->screen_size	 = VFB_SIZE_BYTES;
	info->fix 			 = s1d13c00fb_fix;
	info->fix.mmio_start = (unsigned long)virt_to_phys( (void*)par->RegAddr );
	info->fix.mmio_len	 = par->RegAddrMappedSize;
	info->fix.smem_start = virt_to_phys( (void*)info->screen_base );

	// Initialize Hardware Display Blank
	#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
	par->blank_mode = FB_BLANK_UNBLANK;
	s1d13c00fb_ops.fb_blank = s1d13c00fb_blank;
	#endif

	// Set flags for controller supported features
	info->flags = FBINFO_FLAG_DEFAULT|FBINFO_VIRTFB;

	// Set the pseudo palette
	info->pseudo_palette = par->pseudo_palette;

	s1d13c00fb_checkVar( &info->var, info );
	info->fbops = &s1d13c00fb_ops;
	info->node  = -1;

	info->fbdefio = &s1d13c00fb_defio;
	fb_deferred_io_init( info );

	if ( register_framebuffer(info) < 0 )
	{
		printk( KERN_WARNING "%s(): Could not register frame buffer with kernel.\n", __FUNCTION__ );
		framebuffer_release( info );
		vfree( (void*)info->screen_base );
		return -EINVAL;
	}

	dbg_info( "%s(): fb%d: %s frame buffer device.\n", __FUNCTION__, info->node, info->fix.id );

	#ifdef CONFIG_S1D13C00_IOCTL
	retVal = s1d13c00fb_ioctlSetup( info );
	if ( retVal != 0 )
		return retVal;
	#endif

	#ifdef CONFIG_S1D13C00_PROC
	retVal = s1d13c00fb_procSetup( info );
	#endif

	return retVal;
}




static int s1d13c00fb_remove( struct platform_device* dev )
//
//	Safely shutdown the frame buffer driver.
//
{
	struct fb_info* info = platform_get_drvdata( dev );

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	if ( info )
	{
		#ifdef CONFIG_S1D13C00_PROC
		s1d13c00proc_terminate( );
		#endif

		unregister_framebuffer( info );

		fb_deferred_io_cleanup( info );
		vfree( (void*)info->screen_base );

		framebuffer_release( info );
	}
	return 0;
}




int s1d13c00fb_release( struct fb_info* info, int user )
//
//	Make sure fb updates are performed when /dev/fb? is closed.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	// The S1D13C00 doesn't have a 2D BitBLT Engine, so software BLT.
	cancel_delayed_work( &info->deferred_work );
	schedule_delayed_work( &info->deferred_work, 0 );
	return 0;
}




static int s1d13c00fb_checkVar( struct fb_var_screeninfo* var, struct fb_info* info )
//
//	Fill in the 'var' and 'fix' structures. 
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	var->xres			= C00_DISPLAY_WIDTH;
	var->yres			= C00_DISPLAY_HEIGHT;
	var->xres_virtual	= var->xres;
	var->yres_virtual	= var->yres;
	var->xoffset		= 0;
	var->yoffset		= 0;
	var->bits_per_pixel	= C00_DRIVER_BPP;
	var->grayscale		= 0;
	var->nonstd 		= 0;				// != 0 Non standard pixel format
	var->activate		= FB_ACTIVATE_NOW;	// see FB_ACTIVATE_*
	var->height 		= -1;				// height of picture in mm
	var->width			= -1;				// width of picture in mm
	var->accel_flags	= 0;				// acceleration flags (hints)
	var->pixclock		= 0;				// The LS013B7DD0x panel doesn't really have a pixel clock in the traditional sense
	var->right_margin	= 0;
	var->lower_margin	= 0;
	var->hsync_len		= 0;
	var->vsync_len		= 0;
	var->left_margin	= 0;
	var->upper_margin	= 0;
	var->sync			= 0;
	var->vmode			= FB_VMODE_NONINTERLACED;
	var->red.msb_right 	= var->green.msb_right = var->blue.msb_right = 0;
	var->transp.offset 	= var->transp.length = var->transp.msb_right = 0;

	switch ( info->var.bits_per_pixel )
	{
		case 8:
			var->red.offset = var->green.offset = var->blue.offset = 0;
			var->red.length = var->green.length = var->blue.length = C00_DRIVER_BPP;
			break;

		case 16:
			var->red.offset 	= 11;
			var->red.length 	= 5;
			var->green.offset	= 5;
			var->green.length	= 6;
			var->blue.offset	= 0;
			var->blue.length	= 5;
			var->transp.offset	= 0;
			var->transp.length	= 0;
			break;

		case 24:
			var->red.offset 	= 16;
			var->red.length 	= 8;
			var->green.offset	= 8;
			var->green.length	= 8;
			var->blue.offset	= 0;
			var->blue.length	= 8;
			var->transp.offset	= 0;
			var->transp.length	= 0;
			break;

		default:
			printk( KERN_WARNING "%d BPP is not supported.\n", info->var.bits_per_pixel );
			return -EINVAL;
	}
 	return 0;
}




void s1d13c00fb_fillRect( struct fb_info* info, const struct fb_fillrect* rect )
//
//	Fill rectangle operation.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );
	sys_fillrect( info, rect );
	s1d13c00fb_deferredioTouch( info, (const struct fb_rect*)rect );
}




void s1d13c00fb_copyArea( struct fb_info* info, const struct fb_copyarea* area )
//
//	Copy area operation.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );
	sys_copyarea( info, area );
	s1d13c00fb_deferredioTouch( info, (const struct fb_rect*)area );
}




void s1d13c00fb_imageBlit( struct fb_info* info, const struct fb_image* image )
//
//	Image BLIT operation.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );
	sys_imageblit( info, image );
	s1d13c00fb_deferredioTouch( info, (const struct fb_rect*)image );
}




static void s1d13c00fb_deferredio( struct fb_info* info, struct list_head* pagereflist )
//
//	Copy virtual framebuffer to display framebuffer.
//	Called from the deferred I/O workqueue.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	// It's possible to get here without anything on the pageList via
	// s1d13c00_deferred_io_touch() or via a userspace fsync() invocation
	if ( !list_empty(pagereflist) )
	{
		struct fb_deferred_io_pageref* pageref;
		struct fb_deferred_io_pageref* firstref;
		struct fb_deferred_io*         fbdefio = info->fbdefio;
		int			  				   consecutive_pages = 0;
		int			  				   dirty_pages = 0;
		unsigned long 				   prev_offset = ULONG_MAX;

		// Walk the written page list and copy the data 
		list_for_each_entry( pageref, pagereflist, list )
		{
			dirty_pages++;

			if ( prev_offset == ULONG_MAX )
			{
				firstref = pageref;
				consecutive_pages = 1;
			}
			else if ( (prev_offset+PAGE_SIZE) == pageref->offset )
				consecutive_pages++;
			else
			{
				// Found a non-consecutive page. Update all consecutive pages first...
				s1d13c00fb_update_pages( info, firstref->offset, consecutive_pages );
				firstref = pageref;
				consecutive_pages = 1;
			}
			prev_offset = pageref->offset;
		}

		// Last chunk of consecutive pages
		if ( consecutive_pages )
			s1d13c00fb_update_pages( info, firstref->offset, consecutive_pages );

		// If only a small number of pages were hit, adjust the delay period.
		// This assumes a graphic cursor being moved, so a cursor movement looks less jumpy.
		if ( dirty_pages <= 10 )
			fbdefio->delay = 20;
		else
			fbdefio->delay = VIDEO_REFRESH_PERIOD;
	}
	else
	{
		struct s1d13c00fb_par* par = info->par;

		if ( par->dirtyrange )
		{
			// Process memory region created by s1d13c00_deferred_io_touch()
			s1d13c00fb_flush_dirty_rect( info );
			par->dirtyrange = 0;
		}
		else
		{
			dbg_info( "%s(): NO PAGES.\n", __FUNCTION__ );
			s1d13c00fb_flushVirtualFrameBuffer( info );
		}
	}
}




static void s1d13c00fb_deferredioTouch( struct fb_info* info, const struct fb_rect* rect )
{
	struct fb_deferred_io* fbdefio = info->fbdefio;

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	if ( fbdefio )
	{
		struct s1d13c00fb_par* par = info->par;

		if ( par->dirtyrange == 0 )
		{
			par->dx1 = rect->dx;
			par->dy1 = rect->dy;
			par->dx2 = rect->dx+rect->width;
			par->dy2 = rect->dy+rect->height;
		}
		else
		{
			// Merge the new dirty range with the current one
			if ( rect->dx              < par->dx1 )  par->dx1=rect->dx;
			if ( rect->dy              < par->dy1 )  par->dy1=rect->dy;
			if ( rect->dx+rect->width  > par->dx2 )	 par->dx2=rect->dx+rect->width;
			if ( rect->dy+rect->height > par->dy2 )	 par->dy2=rect->dy+rect->height;
		}

		dbg_info( "%s(): dr:%d dx1:%d dx2:%d dy1:%d dy2:%d.\n", __FUNCTION__, par->dirtyrange, par->dx1, par->dx2, par->dy1, par->dy2 );

		par->dirtyrange = 1;
		schedule_delayed_work( &info->deferred_work, fbdefio->delay );
	}
}




static void s1d13c00fb_update_pages( struct fb_info* info, unsigned long offset, int nPages )
{
	u16* source = (u16*)info->screen_base + offset;

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	#ifdef S1D_EXTRA_DEBUGGING
	printk( KERN_INFO "writeMemSeq offset=%u src=%p np=%d ps=%ld data=%04X %04X %04X %04X\n", offset, source, nPages, PAGE_SIZE, source[0], source[1], source[2], source[3] );
	#endif
	s1d13c00_writeMemBurst( offset, source, nPages*(PAGE_SIZE/2) );
}




static void s1d13c00fb_flush_dirty_rect( struct fb_info* info )
{
	struct s1d13c00fb_par* par = info->par;
	u32					   y, offset, linebytes;
	u32					   stride = C00_DRIVER_STRIDE;
	u16*				   source; 

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	// Optimization: detect the whole screen update
	if ( par->dx1==0 && par->dx2==C00_DISPLAY_WIDTH &&
		 par->dy1==0 && par->dy2==C00_DISPLAY_HEIGHT )
	{
		s1d13c00fb_flushVirtualFrameBuffer( info );
		return;
	}

	offset    = par->dy1*stride + par->dx1*(C00_DRIVER_BPP/8);
	source    = (u16*)info->screen_base + offset;
	linebytes = (par->dx2-par->dx1) * C00_DRIVER_BPP/8;

	for ( y=par->dy1; y<par->dy2; y++ )
	{
		s1d13c00_writeMemBurst( offset, source, linebytes/2 );
		source += stride;
		offset += stride;
	}
}




u16 s1d13c00fb_readMem( struct fb_info* info, u32 offset )
//
//	Reads a pixel from virtual frame buffer.
//
//	offset - Virtual frame buffer offset to start reading from.
//
//	Returns a 16-bit RGB565 value that was read.
//
{
	u16* source = (u16*)info->screen_base;
	return source[offset];
}




void s1d13c00fb_readMemBurst( struct fb_info* info, u32 offset, u16* data, u32 dataLen )
//
//	Reads a series of pixels from virtual frame buffer.
//
//	offset  - Virtual frame buffer offset to start reading.
//	data    - An array containing 16-bit RGB565 pixels that was read from virtual frame buffer.
//	dataLen - Number of pixels to read.
//
{
	u16* source = (u16*)info->screen_base;
	for ( int i=0; i<dataLen; i++ )
		data[i] = source[i+offset];
}



void s1d13c00fb_writeMem( struct fb_info* info, u32 offset, u16 value )
//
//	Writes a 16-bit RGB565 pixel to virtual frame buffer.
//
//	offset - Virtual frame buffer offset to write.
//	value  - The 16-bit RGB565 value to write to virtual frame buffer.
//
{
	u16* source = (u16*)info->screen_base;
	source[offset] = value;
}




void s1d13c00fb_writeMemBurst( struct fb_info* info, u32 offset, const u16* data, u32 dataLen )
//
//	Writes a series of pixels to virtual frame buffer.
//
//	offset  - Virtual frame buffer offset to start writing.
//	data    - An array of 16-bit RGB565 pixels to write to virtual frame buffer.
//	dataLen - Number of pixels to write.
//
{
	u16* source = (u16*)info->screen_base;
	for ( int i=0; i<dataLen; i++ )
		source[i+offset] = data[i];
}




void s1d13c00fb_fillMem( struct fb_info* info, u32 offset, u16 value, u32 len )
//
//	Fills virtual frame buffer with one pixel.
//
//	offset  - Virtual frame buffer offset to start writing.
//	value   - The 16-bit RGB565 value to fill virtual frame buffer.
//	len     - Number of pixels to write.
//
{
	u16* source = (u16*)info->screen_base;
	for ( int i=0; i<len; i++ )
		source[i+offset] = value;
}




void s1d13c00fb_flushVirtualFrameBuffer( struct fb_info* info )
//
//	Flushes the whole virtual framebuffer to the S1D13C00 frame buffer.
//
{
	u32 size   = VFB_SIZE_WORDS;
	u32 offset = 0;

	#ifdef S1D_EXTRA_DEBUGGING
	u16* source = (u16*)info->screen_base;
	printk( KERN_INFO "writeMemSeq offset=%u src=%p n=%u data=%04X %04X %04X %04X\n", offset, source, size, source[0], source[1], source[2], source[3] );
	#endif

	dbg_info( "%s(): offset=%Xh, size=%u.\n", __FUNCTION__, offset, size );
	s1d13c00_writeMemBurst( offset, (u16*)info->screen_base, size );
	s1d13c00_updateDisplay( );
}




#ifdef CONFIG_S1D13C00_DISPLAY_BLANK

int s1d13c00fb_blank( int blank_mode, struct fb_info *info )
//
//	Blank/unblank the display or set the S1D13C00 into PSM (power save mode)
//
{
	struct s1d13c00fb_par* par = info->par;

	dbg_info( "%s(): called with blank_mode=%d.\n", __FUNCTION__, blank_mode );

	// If nothing has changed, just return
	if ( par->blank_mode == blank_mode )
	{
		dbg_info( "%s(): blank_mode=%d (not changed).\n", __FUNCTION__, blank_mode );
		return 0;
	}

	// Switch according to Linux Framebuffer FB_ modes
	switch ( blank_mode )
	{
		case FB_BLANK_UNBLANK:
			// Disable display blanking or disable power save mode
			if ( par->blank_mode == FB_BLANK_POWERDOWN ) 						// If the last mode was PSM, then it is necessary to power up.
				s1d13c00_initialize( false );
			else
				s1d13c00_modifyReg( REG3084_MDCBSTPWR, 0x0001, 0x0001 );		// Enable voltage regulator to panel
			break;

		case FB_BLANK_NORMAL:
			// Blank the display
			s1d13c00_modifyReg( REG3084_MDCBSTPWR, 0x0001, 0x0000 );			// Turn off voltage regulator to panel
			if ( par->blank_mode == FB_BLANK_POWERDOWN )						// If the last mode was PSM, then keep blank_mode as POWERDOWN
				blank_mode = FB_BLANK_POWERDOWN;
			break;

		case FB_BLANK_POWERDOWN:
			// Enter power save mode: turn things off to save power
			s1d13c00_modifyReg( REG3084_MDCBSTPWR, 0x0001, 0x0000 );			// Disable voltage regulator to panel
			s1d13c00_modifyReg( REG30E0_SYSCTRL, 0x0001, 0x0000 );		 		// Disable IOSC
			break;

		default:  // Invalid argument
			dbg_info( "%s(): dropped to default on arg %d.\n", __FUNCTION__, blank_mode );
			return -EINVAL;
	}

	par->blank_mode = blank_mode;
	return 0;
}

#endif // CONFIG_S1D13C00_DISPLAY_BLANK

module_init( s1d13c00fb_init );
module_exit( s1d13c00fb_exit );

MODULE_AUTHOR( "SEIKO EPSON CORPORATION" );
MODULE_DESCRIPTION( "EPSON S1D13C00 Frame Buffer Driver" );
MODULE_LICENSE( "GPL v2" );



