//==============================================================================
//
//	s1d13c00fb.h
//
//	This is the function header for the Epson S1D13C00 frame buffer driver.
//
//	NOTE: All chip-specific features, such as hardware information needed to
//	generate proper timings, color depth, screen resolution, etc., are placed
//	in chip-specific header files and C modules.
//
//	©SEIKO EPSON CORPORATION 2024. All rights reserved.
//
//	This driver software is distributed as is, without any warranty of any
//	kind, either express or implied as further specified in the GNU Public
//	License. This software may be used and distributed according to the terms
//	of the GNU Public License, version 2. See the file COPYING in the main
//	directory of this archive for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.
//
//==============================================================================




#pragma once

#include <linux/kernel.h>
#include <linux/fb.h>
#include <linux/types.h>
#include "s1d13c00_registers.h"	// Include the S1D13C00 register definitions
#include "s1d13c00_settings.h"	// Include the chip specific S1D13C00 register settings
#include "s1d13c00fb_spi.h"
#ifdef CONFIG_S1D13C00_IOCTL
#include "s1d13c00fb_ioctl.h"
#endif




// This is the refresh period for updating the display.
// It is set to (1 second / X).  This value can be modified
// by the end-user.  Be aware that decreasing the refresh period
// increases CPU usage as well. There are HZ (typically 100) jiffies per second
#ifndef CONFIG_S1D13C00_TIMER
#define CONFIG_S1D13C00_TIMER 20
#warning using default CONFIG_S1D13C00_TIMER
#endif

#define VIDEO_REFRESH_PERIOD   (HZ/CONFIG_S1D13C00_TIMER)

// Precalculate the screen dimensions to be used for allocating the virtual buffer
// and clocking the virtual buffer out to the display controller
#define VFB_SIZE_BYTES   (C00_DRIVER_STRIDE*C00_DISPLAY_HEIGHT)
#define VFB_SIZE_WORDS   (VFB_SIZE_BYTES/2)

#ifdef CONFIG_S1D13C00_DEBUG_PRINTK
#define dbg_info(fmt,args...) do { printk(KERN_INFO fmt, ## args); } while (0)
#else
#define dbg_info(fmt,args...) do { } while (0)
#endif




struct s1d13c00fb_par
{
	volatile unsigned char* RegAddr; 
	unsigned				RegAddrMappedSize;
	#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
	int						blank_mode;
	#endif
	u32						pseudo_palette[16];
	int						dirtyrange;
	u32						dx1, dx2, dy1, dy2;
};

extern char*					s1d13c00fb_version;
extern struct fb_fix_screeninfo	s1d13c00fb_fix;




int  s1d13c00fb_init( void );
void s1d13c00fb_exit( void );

u16  s1d13c00fb_readMem( struct fb_info* info, u32 offset );
void s1d13c00fb_readMemBurst( struct fb_info* info, u32 offset, u16* date, u32 dataLen );
void s1d13c00fb_writeMem( struct fb_info* info, u32 offset, u16 value );
void s1d13c00fb_writeMemBurst( struct fb_info* info, u32 offset, const u16* data, u32 dataLen );
void s1d13c00fb_fillMem( struct fb_info* info, u32 offset, u16 value, u32 len );
void s1d13c00fb_flushVirtualFrameBuffer( struct fb_info* info );

#ifdef CONFIG_S1D13C00_IOCTL
int  s1d13c00fb_ioctlSetup( struct fb_info* info );
int  s1d13c00fb_ioctl( struct fb_info *info, unsigned int cmd, unsigned long arg );
#endif

#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
int  s1d13c00fb_blank( int blankMode, struct fb_info* info );
#endif

#ifdef CONFIG_S1D13C00_PROC
int  s1d13c00fb_procSetup( struct fb_info* info );
void s1d13c00proc_terminate( void );
#endif



