//==============================================================================
//
//	s1d13c00fb_ioctl.c
//
//	This file provides the ioctl calls for the Epson S1D13C00 frame buffer
//	driver.
//
//	©SEIKO EPSON CORPORATION 2024. All rights reserved.
//
//	This driver software is distributed as is, without any warranty of any
//	kind, either express or implied as further specified in the GNU Public
//	License. This software may be used and distributed according to the terms
//	of the GNU Public License, version 2. See the file COPYING in the main
//	directory of this archive for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.
//
//==============================================================================




#ifdef CONFIG_S1D13C00_IOCTL




#include <linux/kernel.h>
#include <linux/fb.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/uaccess.h>
#include "s1d13c00fb.h"




static struct fb_info* gs1d13c00_fb = NULL;




int s1d13c00fb_ioctlSetup( struct fb_info* info )
//
//	Sets up the ioctl functions for the S1D13C00 frame buffer driver.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	gs1d13c00_fb = info;

	return 0;
}




int s1d13c00fb_ioctl( struct fb_info* info, unsigned int cmd, unsigned long arg )
//
//	Handles IOCTL messages from the user space.
//
{
	void __user* argp = (void __user*)arg;
	struct s1d13c00_ioctl_hwc ioctl_hwc;

	switch ( cmd )
	{
		case S1D13C00_REGREAD:	// 16-bit read from the S1D13C00 registers
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			ioctl_hwc.value = (int)s1d13c00_readReg( ioctl_hwc.addr );
			dbg_info( "%s(): S1D13C00_REGREAD reg=%04Xh value=%04Xh\n", __FUNCTION__, ioctl_hwc.addr, ioctl_hwc.value );
			return ( copy_to_user(argp,&ioctl_hwc,sizeof(ioctl_hwc)) ? -EFAULT : 0 );

		case S1D13C00_REGWRITE:	// 16-bit write to the S1D13C00 registers
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			dbg_info( "%s(): S1D13C00_REGWRITE reg=%04Xh value=%04Xh\n", __FUNCTION__, ioctl_hwc.addr, ioctl_hwc.value );
			s1d13c00_writeReg( (u32)ioctl_hwc.addr, (u16)ioctl_hwc.value );
			return 0;

		case S1D13C00_REGWRITESEQ: // Multiple 16-bit writes to the S1D13C00 registers
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			dbg_info( "%s(): S1D13C00_REGWRITESEQ value=%04Xh\n", __FUNCTION__, ioctl_hwc.value );
			s1d13c00_writeRegSeq( (u32*)ioctl_hwc.buffer, (u16)ioctl_hwc.value );
			return 0;

		case S1D13C00_MEMREAD:	// 16-bit RGB565 read from the S1D13C00 memory
			// Note: For most cases when the S1D13C00 driver is configured as a virtual frame buffer,
			// the virtual frame buffer should be read from instead of directly from the S1D13C00 memory
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			ioctl_hwc.value = (int)s1d13c00fb_readMem( gs1d13c00_fb, (u32)ioctl_hwc.addr );
			dbg_info( "%s(): S1D13C00_MEMREAD addr=%04Xh value=%02Xh\n", __FUNCTION__, ioctl_hwc.addr, ioctl_hwc.value );
			return ( copy_to_user(argp,&ioctl_hwc,sizeof(ioctl_hwc)) ? -EFAULT : 0 );

		case S1D13C00_MEMWRITE:	// 16-bit RGB565 write to the S1D13C00 memory
			// Note: For most cases when the S1D13C00 driver is configured as a virtual frame buffer,
			// the virtual frame buffer should be written to instead of directly to the S1D13C00 memory
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			dbg_info( "%s(): S1D13C00_MEMWRITE addr=%04Xh value=%02Xh\n", __FUNCTION__, ioctl_hwc.addr, ioctl_hwc.value );
			s1d13c00fb_writeMem( gs1d13c00_fb, (u32)ioctl_hwc.addr, (u16)ioctl_hwc.value );
			return 0;

		case S1D13C00_MEMREADBURST:	// 16-bit RGB565 burst read from the S1D13C00 memory
			// Note: For most cases when the S1D13C00 driver is configured as a virtual frame buffer,
			// the virtual frame buffer should be read from instead of directly from the S1D13C00 memory
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			dbg_info( "%s(): S1D13C00_MEMREADBURST addr=%04Xh value=%02Xh\n", __FUNCTION__, ioctl_hwc.addr, ioctl_hwc.value );
			s1d13c00fb_readMemBurst( gs1d13c00_fb, (u32)ioctl_hwc.addr, (u16*)ioctl_hwc.buffer, (u16)ioctl_hwc.value );
			return ( copy_to_user(argp,&ioctl_hwc,sizeof(ioctl_hwc)) ? -EFAULT : 0 );

		case S1D13C00_MEMWRITEBURST: // 16-bit RGB565 burst write to the S1D13C00 memory
			// Note: For most cases when the S1D13C00 driver is configured as a virtual frame buffer,
			// the virtual frame buffer should be written to instead of directly to the S1D13C00 memory
			if ( copy_from_user(&ioctl_hwc,argp,sizeof(ioctl_hwc)) )
				return -EFAULT;
			dbg_info( "%s(): S1D13C00_MEMWRITEBURST addr=%04Xh value=%02Xh\n", __FUNCTION__, ioctl_hwc.addr, ioctl_hwc.value );
			s1d13c00fb_writeMemBurst( gs1d13c00_fb, (u32)ioctl_hwc.addr, (u16*)ioctl_hwc.buffer, (u16)ioctl_hwc.value );
			return 0;

		case S1D13C00_VBUF_REFRESH:
			dbg_info( "%s(): S1D13C00_VBUF_REFRESH\n", __FUNCTION__ );
			s1d13c00fb_flushVirtualFrameBuffer( info );
			return 0;

		default:
			dbg_info( "%s(): invalid cmd=%xh!\n", __FUNCTION__, cmd );
			return -EINVAL;
	}
}




#endif // CONFIG_S1D13C00_IOCTL



