//==============================================================================
//
//	s1d13c00fb_proc.c
//
//	This file provides the /proc functions for the Epson S1D13C00 frame buffer
//	driver.
//
//	©SEIKO EPSON CORPORATION 2024. All rights reserved.
//
//	This driver software is distributed as is, without any warranty of any
//	kind, either express or implied as further specified in the GNU Public
//	License. This software may be used and distributed according to the terms
//	of the GNU Public License, version 2. See the file COPYING in the main
//	directory of this archive for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.
//
//==============================================================================




#ifdef CONFIG_S1D13C00_PROC


#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/uaccess.h>
#include "s1d13c00_settings.h"
#include "s1d13c00fb.h"




static int     initOpen( struct inode* inode, struct file* file );
static int     initShow( struct seq_file* seq, void* offset );
static int     regOpen( struct inode* inode, struct file* file );
static int     regShow( struct seq_file* seq, void* offset );
static ssize_t regWrite( struct file* file, const char __user* buf, size_t length, loff_t* ppos );
static int     fill16Open( struct inode* inode, struct file* file );
static int     fill16Show( struct seq_file* seq, void* offset );
static ssize_t fill16Write( struct file* file, const char __user* buf, size_t length, loff_t* ppos );
#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
static int     blankOpen( struct inode* inode, struct file* file );
static int     blankShow( struct seq_file* seq, void* offset );
static ssize_t blankWrite( struct file* file, const char __user* buf, size_t length, loff_t* ppos );
#endif
static int     ftOpen( struct inode* inode, struct file* file );
static int     ftShow( struct seq_file* seq, void* offset );
static int     infoOpen( struct inode* inode, struct file* file );
static int     infoShow( struct seq_file* seq, void* offset );
static int     bppOpen( struct inode* inode, struct file* file );
static int     bppShow( struct seq_file* seq, void* offset );
static int     dispTimerOpen( struct inode* inode, struct file* file );
static int     dispTimerShow( struct seq_file* seq, void* offset );




static struct proc_dir_entry* s1d13c00fb_dir;
static struct proc_dir_entry* init_file;
static struct proc_dir_entry* info_file;
static struct proc_dir_entry* ft_file;
static struct proc_dir_entry* reg_file;
static struct proc_dir_entry* fill16_file;
static struct proc_dir_entry* bpp_file;
static struct proc_dir_entry* disptimer_file;
#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
static struct proc_dir_entry* blank_file;
#endif

static u32			   procRegIndex = 0;
static struct fb_info* gs1d13c00_fb = NULL;

static const struct proc_ops s1d13c00fb_init_ops =
{
	.proc_open    = initOpen,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};

static const struct proc_ops s1d13c00fb_info_ops =
{
	.proc_open    = infoOpen,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};
 
static const struct proc_ops s1d13c00fb_reg_ops =
{
	.proc_open    = regOpen,
	.proc_write   = regWrite,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};

static const struct proc_ops s1d13c00fb_fill16_ops =
{
	.proc_open    = fill16Open,
	.proc_write   = fill16Write,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};

#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
static const struct proc_ops s1d13c00fb_blank_ops =
{
	.proc_open    = blankOpen,
	.proc_write   = blankWrite,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};
#endif

static const struct proc_ops s1d13c00fb_ft_ops =
{
	.proc_open    = ftOpen,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};

static const struct proc_ops s1d13c00fb_bpp_ops =
{
	.proc_open    = bppOpen,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};

static const struct proc_ops s1d13c00fb_disptimer_ops =
{
	.proc_open    = dispTimerOpen,
	.proc_read    = seq_read,
	.proc_lseek   = seq_lseek,
	.proc_release = single_release,
};




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/init
//	Provides the init proc function which can be used to initialize the S1D13C00.
//
//	Syntax:
//	cat /proc/s1d13c00fb/init
//
//------------------------------------------------------------------------------




static int initOpen( struct inode* inode, struct file* file )
{
	return single_open( file, initShow, pde_data(inode) );
}




static int initShow( struct seq_file* seq, void* offset )
{
	int error = s1d13c00fb_init( );
	seq_printf( seq, error?"S1D13C00 already initialized.\n":"S1D13C00 initialized.\n" );
	return 0;
}




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/info
//	Provides the info proc function which provides various information,
//	including version information, about the configuration.
//
//	Syntax:
//	cat /proc/s1d13c00fb/info
//
//------------------------------------------------------------------------------




static int infoOpen( struct inode* inode, struct file* file )
{
	return single_open( file, infoShow, pde_data(inode) );
}




static int infoShow( struct seq_file* seq, void* offset )
{
	if ( gs1d13c00_fb == NULL )
	{
		seq_printf( seq, "S1D13C00 is uninitialized!\n\n");
		return 0;
	}

	seq_printf( seq, "%s\n\n", s1d13c00fb_version );
	seq_printf( seq, "Display Resolution: %d x %d\n" "Display Color Depth: %d bpp\n\n", C00_DISPLAY_WIDTH, C00_DISPLAY_HEIGHT, C00_DRIVER_BPP );
	seq_printf( seq, "Virtual Framebuffer Frequency: %d Hz\n", CONFIG_S1D13C00_TIMER );
	if ( gs1d13c00_fb != NULL )
		seq_printf( seq, "Virtual Frame Buffer Addr:%lXh\n", (unsigned long)gs1d13c00_fb->screen_base ); 
	seq_printf( seq, "\nDriver Configuration Details:\n" );

	#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
	seq_printf( seq, "    Hardware Display Blank Enabled\n" );
	#else
	seq_printf( seq, "    Hardware Display Blank Disabled\n" );
	#endif

	#ifdef CONFIG_S1D13C00_PROC
	seq_printf( seq, "    PROC Device Files for Debugging Enabled\n" );
	#else
	seq_printf( seq, "    PROC Device Files for Debugging Disabled\n" );
	#endif

	#ifdef CONFIG_S1D13C00_IOCTL
	seq_printf( seq, "    IOCTL Calls Enabled\n" );
	#else
	seq_printf( seq, "    IOCTL Calls Disbled\n" );
	#endif

	#ifdef CONFIG_S1D13C00_DEBUG_PRINTK
	seq_printf( seq, "    Printk Debugging Messages Enabled\n" );
	#else
	seq_printf( seq, "    Printk Debugging Messages Disabled\n" );
	#endif

	#ifdef CONFIG_S1D13C00_PROC
	seq_printf( seq, "\n"
					 "The PROC functions can be used to read/write S1D13C00 register values:\n\n"
					 "    Read Syntax (register index is hex value):\n"
					 "      echo index=[hex] > /proc/s1d13c00fb/reg\n"
					 "      cat /proc/s1d13c00fb/reg\n\n"
					 "    Write Syntax (register index and val are hex values):\n"
					 "      echo index=AB val=CD > /proc/s1d13c00fb/reg\n\n" );
	#endif
	return 0;
}




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/reg
//	Provides the reg proc function which allows S1D13C00 register IO operations.
//
//	Read Syntax:
//	echo index=hex > /proc/s1d13c00fb/reg
//	cat /proc/s1d13c00fb/reg
//
//	Write Syntax:
//	echo index=hex > /proc/s1d13c00fb/reg
//	echo val=hex > /proc/s1d13c00fb/reg
//	-or-
//	echo index=hex val=hex > /proc/s1d13c00fb/reg
//
//------------------------------------------------------------------------------




static int regOpen( struct inode* inode, struct file* file )
{
	return single_open( file, regShow, pde_data(inode) );
}




static int regShow( struct seq_file* seq, void* offset )
{
	u16 val16 = s1d13c00_readReg( procRegIndex );
	seq_printf( seq, "Register I/O: REG[%04Xh]=%04Xh\n", (u16)procRegIndex, val16 );
	return 0;
}




static ssize_t regWrite( struct file* file, const char __user* buf, size_t length, loff_t* ppos )
{
	char buffer[512];
	uint index, val;

	if ( !buf || length>sizeof(buffer)-1 )
		return -EINVAL;

	if ( copy_from_user(buffer,buf,length) )
		return -EFAULT;

	buffer[length] = '\0';

	if ( sscanf(buffer,"index=%04x val=%04x",&index,&val) == 2 )
	{
		procRegIndex = index;
		s1d13c00_writeReg( procRegIndex, val );
	}
	else if ( sscanf(buffer,"index=%04x",&index) == 1 )
		procRegIndex = index;
	else if ( sscanf(buffer,"val=%x",&val) == 1 )
		s1d13c00_writeReg( procRegIndex, val );
	else
		printk( "%s(): Invalid command %s\n", __FUNCTION__, buffer );
	return length;
}




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/ft
//	Provides the ft proc function which trggers a frame transfer (flush of the
//	virtual	framebuffer.
//
//	Syntax:
//	cat /proc/s1d13c00fb/ft
//
//------------------------------------------------------------------------------




static int ftOpen( struct inode* inode, struct file* file )
{
	return single_open( file, ftShow, pde_data(inode) );
}




static int ftShow( struct seq_file* seq, void* offset )
{
	seq_printf( seq, "Frame transfer.\n" );
	s1d13c00fb_flushVirtualFrameBuffer( gs1d13c00_fb );
	return 0;
}




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/fill16
//	Provides the fill16 proc function which fills the display with a specified
//	byte.
//
//	Syntax:
//	echo XXXXh > /proc/s1d13c00fb/fill16
//
//------------------------------------------------------------------------------




static int fill16Open( struct inode* inode, struct file* file )
{
	return single_open( file, fill16Show, pde_data(inode) );
}




static int fill16Show( struct seq_file* seq, void* offset )
{
	seq_printf( seq, "echo XXh > /proc/s1d13c00fb/fill16\n" );
	seq_printf( seq, "  where XXh is the byte value to fill\n" );
	return 0;
}




static ssize_t fill16Write( struct file* file, const char __user* buf, size_t length, loff_t* ppos )
{
	char buffer[512];
	uint val;

	if ( !buf || length>sizeof(buffer)-1 )
		return -EINVAL;

	if ( copy_from_user(buffer,buf,length) )
		return -EFAULT;

	buffer[length] = '\0';

	if ( sscanf(buffer,"%x",&val) == 1 )
	{
		dbg_info( "%s(): Fill %Xh bytes with %02X.\n", __FUNCTION__, (unsigned)gs1d13c00_fb->screen_size/2, val );
		s1d13c00fb_fillMem( gs1d13c00_fb, 0, (u16)val, C00_DISPLAY_WIDTH*C00_DISPLAY_HEIGHT );
	}
	return length;
}




#ifdef CONFIG_S1D13C00_DISPLAY_BLANK


//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/blank
//	Provides the blank proc function which returns the current "blank"
//	configuration or can be used to set the desired "blank" configuration.
//
//	Read Syntax:
//	cat /proc/s1d13c00fb/blank
//
//	Write Syntax:
//	echo no-blank > /proc/s1d13c00fb/blank
//	-or-
//	echo blank > /proc/s1d13c00fb/blank
//	-or-
//	echo powerdown > /proc/s1d13c00fb/blank
//
//------------------------------------------------------------------------------




static int blankOpen( struct inode* inode, struct file* file )
{
	return single_open( file, blankShow, pde_data(inode) );
}




static int blankShow( struct seq_file* seq, void* offset )
{
	struct s1d13c00fb_par* par = gs1d13c00_fb->par;
	char                   str[50];

	switch ( par->blank_mode )
	{
		case FB_BLANK_UNBLANK:	 strcpy(str,"No Blanking");		 break;
		case FB_BLANK_NORMAL:	 strcpy(str,"Normal Blanking");	 break;
		case FB_BLANK_POWERDOWN: strcpy(str,"Powerdown");		 break;
		default:				 strcpy(str,"Error");			 break;
	}

	seq_printf( seq, "Hardware LCD Blank: %s\nState: %s\n\n", str, par->blank_mode==FB_BLANK_POWERDOWN?"Suspended":"Running" );
	seq_printf( seq, "Change blank state: \necho [no-blank | blank | powerdown] > /proc/s1d13c00fb/blank\n" );
	return 0;
}




static ssize_t blankWrite( struct file* file, const char __user* buf, size_t length, loff_t* ppos )
{
	char buffer[512];

	if ( !buf || length>sizeof(buffer)-1 )
		return -EINVAL;

	if ( copy_from_user(buffer,buf,length) )
		return -EFAULT;

	buffer[length] = '\0';

	if ( !strncmp(buffer,"no-blank",8) )
		s1d13c00fb_blank( FB_BLANK_UNBLANK, gs1d13c00_fb );		// disables display blank and ensures PSM is disabled
	else if ( !strncmp(buffer,"blank",5) )
		s1d13c00fb_blank( FB_BLANK_NORMAL, gs1d13c00_fb );		// enables display blank
	else if ( !strncmp(buffer,"powerdown",9) )
		s1d13c00fb_blank( FB_BLANK_POWERDOWN, gs1d13c00_fb );	// will set into PSM (power save mode)
	else
		printk( "%s(): Invalid command %s\n", __FUNCTION__, buffer );

	return length;
}


#endif // CONFIG_S1D13C00_DISPLAY_BLANK




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/bpp
//	Provides the bpp proc function which provides the current color depth (bpp).
//
//	Syntax:
//	cat /proc/s1d13c00fb/bpp
//
//------------------------------------------------------------------------------




static int bppOpen( struct inode* inode, struct file* file )
{
	return single_open( file, bppShow, pde_data(inode) );
}




static int bppShow( struct seq_file* seq, void* offset )
{
	seq_printf( seq, "Current Color Depth is %d bpp.\n", C00_DRIVER_BPP );
	return 0;
}




//------------------------------------------------------------------------------
//
//	/proc/s1d13c00fb/disptimer
//	Provides the disptimer proc function which displays refresh information for
//	the virtual frame buffer.
//
//	Syntax:
//	cat > /proc/s1d13c00fb/disptimer
//
//------------------------------------------------------------------------------




static int dispTimerOpen( struct inode* inode, struct file* file )
{
	return single_open( file, dispTimerShow, pde_data(inode) );
}




static int dispTimerShow( struct seq_file* seq, void* offset )
{
	seq_printf( seq, "Video refresh period: %d\n", VIDEO_REFRESH_PERIOD );
	seq_printf( seq, "Epson timer divisor: %d\n", CONFIG_S1D13C00_TIMER );
	return 0;
}




//------------------------------------------------------------------------------




int s1d13c00fb_procSetup( struct fb_info* info )
//
//	Sets up the proc functions for the S1D13C00 frame buffer driver.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	gs1d13c00_fb = info;

	// First setup a subdirectory for s1d13c00fb
	s1d13c00fb_dir = proc_mkdir( "s1d13c00fb", NULL );

	if ( !s1d13c00fb_dir )
		return -ENOMEM;

	/* 'init' [R] */
	init_file = proc_create_data( "init", S_IRUGO, s1d13c00fb_dir, &s1d13c00fb_init_ops, NULL );
	if ( init_file == NULL )
		goto err_remove_post_info;
 
	/* 'info' [R] */
	info_file = proc_create_data( "info", S_IRUGO, s1d13c00fb_dir, &s1d13c00fb_info_ops, NULL );
	if ( info_file == NULL )
		goto err_remove_post_info;

	/* 'ft' [R] */
	ft_file = proc_create_data( "ft", S_IRUGO, s1d13c00fb_dir, &s1d13c00fb_ft_ops, NULL );
	if ( ft_file == NULL )
		goto err_remove_post_info;

	/* 'reg' [R/W] */
	reg_file = proc_create_data( "reg", S_IFREG|S_IRUGO|S_IWUSR, s1d13c00fb_dir, &s1d13c00fb_reg_ops, NULL );
	if ( reg_file == NULL )
		goto err_remove_post_info;

	/* 'fill16' [R/W] */
	fill16_file = proc_create_data( "fill16", S_IFREG|S_IRUGO|S_IWUSR, s1d13c00fb_dir, &s1d13c00fb_fill16_ops, NULL );
	if ( fill16_file == NULL )
		goto err_remove_post_info;

	#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
	/* 'blank' [R/W] */
	blank_file = proc_create_data( "blank", S_IFREG|S_IRUGO|S_IWUSR, s1d13c00fb_dir, &s1d13c00fb_blank_ops, NULL );
	if ( blank_file == NULL )
		goto err_remove_post_info;
	#endif
	
	/* 'bpp' [R] */
	bpp_file = proc_create_data( "bpp", S_IRUGO, s1d13c00fb_dir, &s1d13c00fb_bpp_ops, NULL );
	if ( bpp_file == NULL )
		goto err_remove_post_info;

	/* 'disptimer' [R] */
	disptimer_file = proc_create_data( "disptimer", S_IRUGO, s1d13c00fb_dir, &s1d13c00fb_disptimer_ops, NULL );
	if ( disptimer_file == NULL )
		goto err_remove_post_info;

	return 0;

err_remove_post_info:
	s1d13c00proc_terminate( );
	return -ENOMEM;
}




void s1d13c00proc_terminate( void )
//
//	Terminates the proc functions for the S1D13C00 frame buffer driver and cleans up.
//
{
	dbg_info( "%s(): called.\n", __FUNCTION__ );

	if ( s1d13c00fb_dir )
	{
		if (init_file)		remove_proc_entry("init",s1d13c00fb_dir);
		if (info_file)		remove_proc_entry("info",s1d13c00fb_dir);
		if (ft_file)		remove_proc_entry("ft",s1d13c00fb_dir);
		if (reg_file)		remove_proc_entry("reg",s1d13c00fb_dir);
		if (fill16_file)	remove_proc_entry("fill16",s1d13c00fb_dir);
		if (bpp_file)		remove_proc_entry("bpp",s1d13c00fb_dir);
		if (disptimer_file)	remove_proc_entry("disptimer",s1d13c00fb_dir);

		#ifdef CONFIG_S1D13C00_DISPLAY_BLANK
		if ( blank_file )
		{
			remove_proc_entry( "blank", s1d13c00fb_dir );
			blank_file = NULL;
		}
		#endif

		remove_proc_entry( "s1d13c00fb", NULL );
	}

	s1d13c00fb_dir = init_file = ft_file = info_file = reg_file = fill16_file = bpp_file = disptimer_file = NULL;
}


#endif /* CONFIG_S1D13C00_PROC */



