//==============================================================================
//
//	s1d13c00fb_spi.c
//
//	Interface code for implementing an interface between the Raspberry Pi
//	and the Epson S1D13C00.
//
//	©SEIKO EPSON CORPORATION 2024. All rights reserved.
//
//	This driver software is distributed as is, without any warranty of any
//	kind, either express or implied as further specified in the GNU Public
//	License. This software may be used and distributed according to the terms
//	of the GNU Public License, version 2. See the file COPYING in the main
//	directory of this archive for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.
//
//==============================================================================




#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/fb.h>
#include <linux/spi/spi.h>
#include "s1d13c00fb.h"
#include "s1d13c00fb_spi.h"




// Local functions
static void s1d13c00_allocBuffers( int len );
static void s1d13c00_freeBuffers( void );
static bool s1d13c00_readWriteSPI( u8* sendBuf, u8* recvBuf, int len );
static bool s1d13c00_writeSPI( u8* sendBuf, u8* recvBuf, int len );
static int  s1d13c00_spi_probe( struct spi_device* spi );


// Local definitions
static struct spi_device* spiDevice = NULL;
static u8*    tx = NULL;
static u8*    rx = NULL;

#define NA	0  // Not Applicable

static u32 initSeq[] =		 // Register initialization sequence for the LS013B7DD0x panel
{
	// REGISTER			 	 MASK    VALUE

	// Soft Reset.
	REG30E0_SYSCTRL,		 0xFFFF, 0x8000,
	REGFLAG_WAIT_MS, 		 NA,     1,

	// Start IOSC.
	REG30E0_SYSCTRL,		 0xFFFF, 0x11|(3<<1)|(39<<8),							// IOSCFQ=20MHz, IOSC Trim Adjust Value=39
	REGFLAG_WAIT_IOSC,		 NA,     0,

	// Start OSC1.
	REG0000_SYSPROT,		 0xFFFF, 0x0096,										// Write protect off
	REG0046_CLGOSC1,		 0xFFFF, 0x5092,
	REG0000_SYSPROT,		 0xFFFF, 0x0000,										// Write protect on
	REG004C_CLGINTF,		 0xFFFF, 0x00FF,										// Clear interrupts
	REG0042_CLGOSC,			 0xFFFF, 0x0002,										// Enable OSC1
	REGFLAG_WAIT_OSC1,		 NA,     0,

	// Sequence power-on of voltage supplies.
	REG3080_MDCBSTCLK,		 0xFFFF, (5<<4)|1,
	REG3084_MDCBSTPWR,		 0xFFFF, 0x0007,										// VMDBUP=normal, BSTON=on, REGECO=on, REGON=on
	REGFLAG_WAIT_MS, 		 NA,     2,

	// T1: Turn on VMDL.
	REG3088_MDCBSTVMD,		 0xFFFF, (2<<12)|(4<<4)|0x0001,							// VMDH=4.5V, VMDL=3.2V
	REGFLAG_WAIT_MS,		 NA,     2,

	// T2: Turn on VMDH.
	REG3088_MDCBSTVMD,		 0xFFFF, (2<<12)|(4<<4)|0x0101,							// VMDH=4.5V, VMDL=3.2V
	REGFLAG_WAIT_MS,		 NA,     2,

	// T3: Panel internal circuit reset time.
	REGFLAG_WAIT_MS,		 NA,     1,

	// T4: Initialize clocks and panel interface.
	REG3002_MDCDISPWIDTH,	 0xFFFF, C00_DISPLAY_WIDTH,
	REG3004_MDCDISPHEIGHT,	 0xFFFF, C00_DISPLAY_HEIGHT,
	REG3016_MDCDISPSTRIDE,	 0xFFFF, C00_DISPLAY_STRIDE,
	REG3006_MDCDISPVCOMDIV,	 0xFFFF, 136,											// VCOM freq=32768/(4*(136+1))=~60Hz
	REG3008_MDCDISPCLKDIV,	 0xFFFF, 0x0100|(((u16)(C00_IOSC_FREQ/2850000UL))-1),	// Divider value to get T=0.35us(2.85MHz), VCKHST=(1+1)*T=0.7us
	REG300A_MDCDISPPRM21,	 0xFFFF, 0x0080,										// HSTHCK=0x00=(0+1)*T=0.35us, VSTVCK=0x80=(128+1)*T=45.15us
	REG300C_MDCDISPPRM43,	 0xFFFF, 0x1900,										// VCK to ENB:t0+t2+25*2T=0.7+0.35+25*0.7=18.55us, HCKDLY=(0+1)*T=0.35us
	REG300E_MDCDISPPRM65,	 0xFFFF, 0xFF27,										// XRSTVST=(255+1)*T=89.6us, ENBW=39 HCK counts=39*0.7us=27.3us
	REG3010_MDCDISPPRM87,	 0xFFFF, 0x0000,										// End of line to VCK=(0+1)*2T=0.35us, HCK to HST/data=(0+1)*T=0.35us
	REG3054_MDCDISPPRM109,	 0xFFFF, 0x0301,										// Extra HCK counts after end of line pixel data=3, HCK count for start of pixel data=1
	REG3056_MDCDISPPRM1211,	 0xFFFF, 0x0302,										// Extra VCK counts after last line of pixel data, VCK count for start of pixel data
	REG3058_MDCDISPPRM1413,	 0xFFFF, 0x0003,										// Fast VCK high/low width=3*T=1.05us
	REG3018_MDCDISPFRMBUFF0, 0xFFFF, C00_MEM_START_ADDR&0xFFFF,
	REG301A_MDCDISPFRMBUFF1, 0xFFFF, C00_MEM_START_ADDR>>16,
	REG3000_MDCDISPCTL,		 0xFFFF, 0x0000,										// Initially VCOM should be disabled
	REG305A_MDCDISPCTL2,	 0xFFFF, 0x0001,
	REG3044_MDCGFXOBADDR0,	 0xFFFF, C00_MEM_START_ADDR&0xFFFF,
	REG3046_MDCGFXOBADDR1,	 0xFFFF, C00_MEM_START_ADDR>>16,
	REG302E_MDCGFXOWIDTH,	 0xFFFF, C00_DISPLAY_WIDTH,
	REG3030_MDCGFXOHEIGHT,	 0xFFFF, C00_DISPLAY_HEIGHT,
	REG304A_MDCGFXOSTRIDE,	 0xFFFF, C00_DISPLAY_STRIDE,
};
static int initSeqLen = sizeof(initSeq) / sizeof(initSeq[0]);

static u32 clearSeq[] =		 // Clear the LS013B7DD0x panel with a black rectangle.
{
	// REGISTER			 	 MASK    VALUE
	REG3022_MDCGFXIXCENTER,	 0xFFFF, 0,												// Top-left corner X
	REG3024_MDCGFXIYCENTER,	 0xFFFF, 0,												// Top-left corner Y
	REG302A_MDCGFXOXCENTER,	 0xFFFF, C00_DISPLAY_WIDTH-1,							// Bottom-right corner X
	REG302C_MDCGFXOYCENTER,	 0xFFFF, C00_DISPLAY_HEIGHT-1,							// Bottom-right corner Y
	REG303E_MDCGFXCOLOR,	 0xFFFF, C00_DISPLAY_BG_COLOR,							// Background color (00000000AARRGGBB)
	REG3026_MDCGFXIWIDTH,	 0xFFFF, 0,												// V Line Thickness
	REG3028_MDCGFXIHEIGHT,	 0xFFFF, 0,												// H Line Thickness
	REG3020_MDCGFXCTL,		 0xFFFF, (1<<11),										// Enable fill
	REG301E_MDCINTCTL,		 0x0001, 0x0001,										// Clear interrupt flag
	REG3020_MDCGFXCTL,		 0x0007, 0x0002,										// RECTDRAW
	REG301C_MDCTRIGCTL,		 0x0001, 0x0001,										// Trigger RECTDRAW
	REGFLAG_WAIT_GFX,		 NA,     NA,											// Wait until rectangle is drawn
};
static int clearSeqLen = sizeof(clearSeq) / sizeof(clearSeq[0]);

static u32 updateSeq[] =	 // Tigger update display.
{
	// REGISTER			 	 MASK    VALUE
	REG3012_MDCDISPSTARTY,	 0xFFFF, 0,
	REG3014_MDCDISPENDY,	 0xFFFF, C00_DISPLAY_HEIGHT-1,
	REGFLAG_UPDATE_DISPLAY,	 NA,     NA,											// Wait for display update
	REG3000_MDCDISPCTL,		 0x0002, 0x0002,										// Enable VCOM
	REGFLAG_WAIT_MS,		 NA,     17
};
static int updateSeqLen = sizeof(updateSeq) / sizeof(updateSeq[0]);




static int s1d13c00_spi_probe( struct spi_device* spi )
//
//	Entry point for SPI driver (this display driver also acts as a SPI driver).
//
{
	int result;
	u16 prodCode;

	dbg_info( "%s(): called.\n", __FUNCTION__ );

	spiDevice = spi;
	spiDevice->mode = SPI_MODE_0;
	spiDevice->bits_per_word = 8;
	result = spi_setup( spiDevice );
	if ( result < 0 )
	{
		printk( KERN_WARNING "%s(): ERROR! spi_setup() failed\n", __FUNCTION__ );
		return false;
	}

	prodCode = s1d13c00_readReg( REG306C_PRODCODE );
	if ( prodCode != 0x005F )
	{
		printk( KERN_WARNING "%s(): ERROR! Product Code is not 005Fh (REG[%04Xh]==%04Xh).\n", __FUNCTION__, REG306C_PRODCODE&0x0000FFFF, prodCode );
		return false;
	}
	dbg_info( "%s(): S1D13C00 detected.\n", __FUNCTION__ );

	s1d13c00_initialize( true );

	return 0;
}




void s1d13c00_initialize( bool clearScreen )
//
//	Resets and re-initializes the S1D13C00 chip.
//
{
	dbg_info( "%s(): Write the 'init' register sequence:\n", __FUNCTION__ );
	s1d13c00_writeRegSeq( initSeq, initSeqLen );
	if ( clearScreen )
	{
		dbg_info( "%s(): Write the 'clear screen' register sequence:\n", __FUNCTION__ );
		s1d13c00_writeRegSeq( clearSeq, clearSeqLen );
	}
	dbg_info( "%s(): Write the 'update display' register sequence:\n", __FUNCTION__ );
	s1d13c00_writeRegSeq( updateSeq, updateSeqLen );
}




void s1d13c00_updateDisplay( void )
//
//	Triggers C00 to update its display.
//
{
	s1d13c00_modifyReg( REG301E_MDCINTCTL, 0x0002, 0x0002 );	// Clear UPDINT flag
	s1d13c00_modifyReg( REG301C_MDCTRIGCTL, 0x0002, 0x0002 );	// Trigger display update
	s1d13c00_modifyReg( REG301E_MDCINTCTL, 0x0202, 0x0200 );	// Enable interrupt
	while ( (s1d13c00_readReg(REG30E4_SYSINTS)&0x0010) == 0 )	// Poll wait until MDC interrupt occurs
		continue;
	s1d13c00_modifyReg( REG301E_MDCINTCTL, 0x0002, 0x0002 );	// Clear interrupt
}




void s1d13c00_allocBuffers( int len )
//
//	Ensures transmit and recieve buffers are allocated with enough memory.
//
{
	tx = kcalloc( len, sizeof(u8), GFP_KERNEL );
	rx = kcalloc( len, sizeof(u8), GFP_KERNEL );
}




void s1d13c00_freeBuffers( void )
//
//	Frees transmit and recieve buffers.
//
{
	if ( tx != NULL )
	{
		kfree( tx );
		tx = NULL;
	}
	if ( rx != NULL )
	{
		kfree( rx );
		rx = NULL;
	}
}




bool s1d13c00_readWriteSPI( u8* sendBuf, u8* recvBuf, int len )
//
//	Sends a bi-directional SPI sequence.
//
//	Returns true if successful.
//
{
	int retVal;
	struct spi_transfer xfer;
	memset( &xfer, 0, sizeof(xfer) );

	// NOTE: Read speed needs to slow down to 1 MHz while reading.  However,
	// it is currently set to 600 KHz due to the specific daughter board being
	// used. Feel free to set it to 1 MHz if your platform can handle it.
	xfer.speed_hz = 600000; // 600 KHz
	xfer.tx_buf = sendBuf;
	xfer.rx_buf = recvBuf;
	xfer.len = len;

	retVal = spi_sync_transfer( spiDevice, &xfer, 1 );
	return ( retVal == 0 );
}




bool s1d13c00_writeSPI( u8* sendBuf, u8* recvBuf, int len )
//
//	Sends a SPI sequence that performs a write only.
//
//	Returns true if successful.
//
{
	int retVal;
	struct spi_transfer xfer;
	memset( &xfer, 0, sizeof(xfer) );

	// NOTE: Writes can be set to a faster speed if not reading.
	xfer.speed_hz = 5700000; // 5.7 MHz
	xfer.tx_buf = sendBuf;
	xfer.rx_buf = recvBuf;
	xfer.len = len;

	retVal = spi_sync_transfer( spiDevice, &xfer, 1 );
	return ( retVal == 0 );
}



u16 s1d13c00_readReg( u32 index )
//
//	Reads a register value.
//
//	index - Register index.
//
//	Returns register value.
//
{
	u32 addr = C00_REG_START_ADDR | index;
	u16 value;

	s1d13c00_allocBuffers( 8 );
	tx[0] = 0x03; // "single read" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	tx[5] = 0x00;
	tx[6] = 0x00;
	tx[7] = 0x00;

	if ( s1d13c00_readWriteSPI(tx,rx,8) )
		value = ((u16)rx[7]<<8) | rx[6];
	else
		value = 0xEEEE;

	s1d13c00_freeBuffers( );
	return value;
}




void s1d13c00_writeReg( u32 index, u16 value )
//
//	Writes a value to a register.
//
//	index - Register index.
//	value - New value.
//
{
	u32 addr = C00_REG_START_ADDR | index;

	dbg_info( "%s():    REG[%04Xh]=%04Xh\n", __FUNCTION__, index&0x0000FFFF, value );

	s1d13c00_allocBuffers( 7 );
	tx[0] = 0x02; // "single write" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	tx[5] = (u8)( value & 0xFF );
	tx[6] = (u8)( value >> 8 );

	s1d13c00_writeSPI( tx, rx, 7 );
	s1d13c00_freeBuffers( );
}




void s1d13c00_writeRegSeq( u32* seq, int seqLen )
//
//	Writes multiple registers in sequence.
//
//	seq    - A sequence containing register index and value pairs.
//	seqLen - Total count of all register indexes and values.
//
{
	for ( int i=0; i<seqLen; i+=3 )
	{
		switch( seq[i] )
		{
			case REGFLAG_WAIT_MS:
				dbg_info( "%s(): WAIT %dms\n", __FUNCTION__, seq[i+2] );
				msleep( seq[i+2] );
				break;

			case REGFLAG_WAIT_IOSC:
				dbg_info( "%s(): WAIT IOSC\n", __FUNCTION__ );
				while ( (s1d13c00_readReg(REG30E0_SYSCTRL)&0x0008) == 0 )	// Wait until IOSC is stable
					continue;
				break;

			case REGFLAG_WAIT_OSC1:
				dbg_info( "%s(): WAIT OSC1\n", __FUNCTION__ );
				while ( (s1d13c00_readReg(REG004C_CLGINTF)&0x0002) == 0 )	// Wait until OSC1 is stable
					continue;
				break;

			case REGFLAG_WAIT_GFX:
				dbg_info( "%s(): WAIT GFX\n", __FUNCTION__ );
				s1d13c00_modifyReg( REG301E_MDCINTCTL, 0x0101, 0x0100 );	// Enable interrupt
				while ( (s1d13c00_readReg(REG30E4_SYSINTS)&0x0010) == 0 )	// Poll wait until MDC interrupt occurs
					continue;
				s1d13c00_modifyReg( REG301E_MDCINTCTL, 0x0001, 0x0001 );	// Clear interrupt
				break;

			case REGFLAG_UPDATE_DISPLAY:
				dbg_info( "%s(): UPDATE DISPLAY\n", __FUNCTION__ );
				s1d13c00_updateDisplay( );
				break;

			default:
				s1d13c00_modifyReg( (u32)seq[i], (u16)seq[i+1], (u16)seq[i+2] );
				break;
		}
	}
}




u16 s1d13c00_modifyReg( u32 index, u16 mask, u16 value )
//
//	Modifies only certain bits of a register.
//
//	index - Register index.
//	mask  - Only "on" (1) bits will be changed.
//	value - New value (but the register will be written with only bits that correspond to the mask).
//
//	Returns the new register value.
//
{
	u16 newValue = value & mask;

	if ( mask != 0xFFFF )
		newValue |= s1d13c00_readReg(index) & ~mask;
	s1d13c00_writeReg( index, newValue );
	return newValue;
}




u16 s1d13c00_readMem( u32 offset )
//
//	Reads a pixel from display memory.
//
//	offset - Display memory offset to start reading from.
//
//	Returns a 16-bit RGB565 value that was read.
//
{
	u32 addr = C00_MEM_START_ADDR | offset;
	u16 value;

	s1d13c00_allocBuffers( 7 );
	tx[0] = 0x03; // "single read" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	tx[5] = 0x00;
	tx[6] = 0x00;

	if ( s1d13c00_readWriteSPI(tx,rx,7) )
		value = TO_16_BPP( rx[6] );
	else
		value = 0xEEEE;
	s1d13c00_freeBuffers( );
	return value;
}




void s1d13c00_readMemBurst( u32 offset, u16* data, u32 dataLen )
//
//	Reads a series of pixels from display memory.
//
//	offset  - Display memory offset to start reading.
//	data    - An array containing 16-bit RGB565 pixels that was read from display memory.
//	dataLen - Number of pixels to read.
//
{
	u32 addr = C00_MEM_START_ADDR | offset;
	u32 header = 6;

	s1d13c00_allocBuffers( header+dataLen );
	tx[0] = 0x03; // "single read" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	tx[5] = 0x00;
	for ( u32 i=0; i<dataLen; i++ )
		tx[header+i] = 0x00;

	if ( s1d13c00_readWriteSPI(tx,rx,header+dataLen) )
		for ( u32 i=0; i<dataLen; i++ )
			data[i] = TO_16_BPP( rx[header+i] );
	s1d13c00_freeBuffers( );
}



void s1d13c00_writeMem( u32 offset, u16 value )
//
//	Writes a 16-bit RGB565 pixel to display memory.
//
//	offset - Display memory offset to write.
//	value  - The 16-bit RGB565 value to write to display memory.
//
{
	u32 addr = C00_MEM_START_ADDR | offset;

	s1d13c00_allocBuffers( 6 );
	tx[0] = 0x02; // "single write" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	tx[5] = TO_8_BPP( value );

	s1d13c00_writeSPI( tx, rx, 6 );
	s1d13c00_freeBuffers( );
}




void s1d13c00_writeMemBurst( u32 offset, const u16* data, u32 dataLen )
//
//	Writes a series of pixels to display memory.
//
//	offset  - Display memory offset to start writing.
//	data    - An array of 16-bit RGB565 pixels to write to display memory.
//	dataLen - Number of pixels to write.
//
{
	u32 addr = C00_MEM_START_ADDR | offset;
	u32 header = 5;

	s1d13c00_allocBuffers( header+dataLen );
	tx[0] = 0x02; // "single write" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	for ( u32 i=0; i<dataLen; i++ )
		tx[header+i] = TO_8_BPP( data[i] );

	s1d13c00_writeSPI( tx, rx, header+dataLen );
	s1d13c00_freeBuffers( );
}




void s1d13c00_fillMem( u32 offset, u16 value, u32 len )
//
//	Fills display memory with one pixel.
//
//	offset  - Display memory offset to start writing.
//	value   - The 16-bit RGB565 value to fill display memory.
//	len     - Number of pixels to write.
//
{
	u8  actualColor = TO_8_BPP( value );
	u32 addr = C00_MEM_START_ADDR | offset;
	u32 header = 5;

	s1d13c00_allocBuffers( header+len );
	tx[0] = 0x02; // "single write" command
	tx[1] = (u8)( (addr>>24) & 0xFF );
	tx[2] = (u8)( (addr>>16) & 0xFF );
	tx[3] = (u8)( (addr>>8) & 0xFF );
	tx[4] = (u8)( addr & 0xFF );
	for ( u32 i=0; i<len; i++ )
		tx[header+i] = actualColor;

	s1d13c00_writeSPI( tx, rx, header+len );
	s1d13c00_freeBuffers( );
}




static const struct of_device_id s1d13c00_of_match[] =
{
	{ .compatible = "epson,s1d13c00" },
	{ },
};
MODULE_DEVICE_TABLE( of, s1d13c00_of_match );


static const struct spi_device_id s1d13c00_spi_ids[] =
{
	{ "s1d13c00", 0 },
	{ }
};
MODULE_DEVICE_TABLE( spi, s1d13c00_spi_ids );


static struct spi_driver s1d13c00_spi_driver =
{
	.probe = s1d13c00_spi_probe,
	.id_table = s1d13c00_spi_ids,
	.driver =
	{
		.name  = "s1d13c00",
		.owner = THIS_MODULE,
		.of_match_table = of_match_ptr(s1d13c00_of_match),
	},
};
module_spi_driver( s1d13c00_spi_driver );



