//========================================================================================================
//
//	testbed_with_driver.c - A test app that accesses C00 using the C00's display driver on a Raspberry Pi.
//
//	This test app can be used to communicate to C00 after the C00's display driver is installed.
//
//	©SEIKO EPSON CORPORATION 2024. All rights reserved.
//
//	This driver software is distributed as is, without any warranty of any
//	kind, either express or implied as further specified in the GNU Public
//	License. This software may be used and distributed according to the terms
//	of the GNU Public License, version 2. See the file COPYING in the main
//	directory of this archive for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.
//
//========================================================================================================




#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/spi/spidev.h>
#include "../s1d13c00_registers.h"
#include "../s1d13c00_settings.h"
#include "../s1d13c00fb_ioctl.h"




typedef uint8_t  u8;
typedef uint16_t u16;
typedef uint32_t u32;

#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))


// Local functions
static bool s1d13c00_init( );
static void s1d13c00_exit( );
static void s1d13c00_checkBufferSize( int len );
static u16  s1d13c00_readReg( u32 index );
static void s1d13c00_writeReg( u32 index, u16 value );
static void s1d13c00_writeRegSeq( u32* seq, int seqLen );
static u16  s1d13c00_modifyReg( u32 index, u16 mask, u16 value );
static u16  s1d13c00_readMem( u32 offset );
static void s1d13c00_readMemBurst( u32 offset, u16* data, u32 dataLen );
static void s1d13c00_writeMem( u32 offset, u16 value );
static void s1d13c00_writeMemBurst( u32 offset, u16* data, u32 dataLen );
static void s1d13c00_updateDisplay( );
static u16  readPixel( int x, int y );
static void writePixel( int x, int y, u16 color );
static void writeRect( int x, int y, int width, int height, u16 color );
static long getTime( );
static void printTimeDiff( long startTime );


// Local definitions
#define BLACK					0xC0
#define RED						0xC000
#define GREEN					0x0400
#define BLUE					0x0018
#define MAGENTA					(RED|BLUE)

static int dev = -1;
static struct s1d13c00_ioctl_hwc ioctl_hwc;




bool s1d13c00_init( )
//
//	Initializes the usage of S1D13C00.
//
{
	dev = open( "/dev/fb1", O_RDWR );	// <-- NOTE: Currently using fb1, please change this if C00 is on fb0.
	if ( dev < 0 )
	{
		printf( "ERROR: Cannot open spi device!\n" );
		return false;
	}

	return true;
}




void s1d13c00_exit( )
//
//	Terminates usage of the S1D13C00.
//
{
	if ( dev >= 0 )
	{
		close( dev );
		dev = -1;
	}
}




u16 s1d13c00_readReg( u32 index )
//
//	Reads a register value.
//
//	index - Register index.
//
//	Returns register value.
//
{
	ioctl_hwc.addr = C00_REG_START_ADDR | index;
	int error = ioctl( dev, S1D13C00_REGREAD, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_REGREAD IOCTL error: %d!\n", error );
	return ioctl_hwc.value;
}




void s1d13c00_writeReg( u32 index, u16 value )
//
//	Writes a value to a register.
//
//	index - Register index.
//	value - New value.
//
{
	ioctl_hwc.addr  = C00_REG_START_ADDR | index;
	ioctl_hwc.value = value;
	int error = ioctl( dev, S1D13C00_REGWRITE, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_REGWRITE IOCTL error: %d!\n", error );
}




void s1d13c00_writeRegSeq( u32* seq, int seqLen )
//
//	Writes multiple registers in sequence.
//
//	seq    - A sequence containing register index and value pairs.
//	seqLen - Total count of all register indexes and values.
//
{
	ioctl_hwc.buffer = seq;
	ioctl_hwc.value  = seqLen;
	int error = ioctl( dev, S1D13C00_REGWRITESEQ, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_REGWRITESEQ IOCTL error: %d!\n", error );
}




u16 s1d13c00_modifyReg( u32 index, u16 mask, u16 value )
//
//	Modifies only certain bits of a register.
//
//	index - Register index.
//	mask  - Only "on" (1) bits will be changed.
//	value - New value (but the register will be written with only bits that correspond to the mask).
//
//	Returns the new register value.
//
{
	u16 newValue = value & mask;

	if ( mask != 0xFFFF )
		newValue |= s1d13c00_readReg(index) & ~mask;
	s1d13c00_writeReg( index, newValue );
	return newValue;
}




u16 s1d13c00_readMem( u32 offset )
//
//	Reads a pixel from display memory.
//
//	offset - Display memory offset to start reading from.
//
//	Returns a 16-bit RGB565 value that was read.
//
{
	ioctl_hwc.addr = offset;
	int error = ioctl( dev, S1D13C00_MEMREAD, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_MEMREAD IOCTL error: %d!\n", error );
	return (u16)ioctl_hwc.value;
}




void s1d13c00_readMemBurst( u32 offset, u16* data, u32 dataLen )
//
//	Reads a series of pixels from display memory.
//
//	offset  - Display memory offset to start reading.
//	data    - An array containing 16-bit RGB565 pixels that was read from display memory.
//	dataLen - Number of pixels to read.
//
{
	ioctl_hwc.addr   = offset;
	ioctl_hwc.buffer = data;
	ioctl_hwc.value  = dataLen;
	int error = ioctl( dev, S1D13C00_MEMREADBURST, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_MEMREADBURST IOCTL error: %d!\n", error );
}



void s1d13c00_writeMem( u32 offset, u16 value )
//
//	Writes a 16-bit RGB565 pixel to display memory.
//
//	offset - Display memory offset to write.
//	value  - The 16-bit RGB565 value to write to display memory.
//
{
	ioctl_hwc.addr  = offset;
	ioctl_hwc.value = value;
	int error = ioctl( dev, S1D13C00_MEMWRITE, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_MEMWRITE IOCTL error: %d!\n", error );
}




void s1d13c00_writeMemBurst( u32 offset, u16* data, u32 dataLen )
//
//	Writes a series of pixels to display memory.
//
//	offset  - Display memory offset to start writing.
//	data    - An array of 16-bit RGB565 pixels to write to display memory.
//	dataLen - Number of pixels to write.
//
{
	ioctl_hwc.addr   = offset;
	ioctl_hwc.buffer = data;
	ioctl_hwc.value  = dataLen;
	int error = ioctl( dev, S1D13C00_MEMWRITEBURST, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_MEMWRITEBURST IOCTL error: %d!\n", error );
}




void s1d13c00_updateDisplay( )
//
//	Triggers C00 to update its display.
//
{
	int error = ioctl( dev, S1D13C00_VBUF_REFRESH, &ioctl_hwc );
	if ( error != 0 )
		printf( "S1D13C00_VBUF_REFRESH IOCTL error: %d!\n", error );
}




u16 readPixel( int x, int y )
//
//	Reads a pixel from the display.
//
//	Returns the pixel color.
//
{
	return s1d13c00_readMem( (y*C00_DISPLAY_STRIDE)+x );
}




void writePixel( int x, int y, u16 color )
//
//	Writes a pixel to the display.
//
{
	s1d13c00_writeMem( (y*C00_DISPLAY_STRIDE)+x, color );
}




void writeRect( int x, int y, int width, int height, u16 color )
//
//	Writes a solid rectangle to the display.
//
{
	u16* data = calloc( width, sizeof(u16) );
	for ( int xx=0; xx<width; xx++ )
		data[xx] = color;
	for ( int yy=0; yy<height; yy++ )
		s1d13c00_writeMemBurst( (y+yy)*C00_DISPLAY_STRIDE+x, data, width );
	free( data );
}




long getTime( )
//
//	Gets system time.
//	Used with printTimeDiff() for benchmarking performance.
//
{
	struct timeval t;
	gettimeofday( &t, NULL );
	return (long)t.tv_sec*1000 + (long)t.tv_usec/1000;
}




void printTimeDiff( long startTime )
//
//	Shows system time duration since a call to getTime().
//	Used with getTime() for benchmarking performance.
//
{
	printf( "Time took %.2f seconds\n", (double)(getTime()-startTime)/1000 );
}




int main( )
{
	u32 value;

	if ( s1d13c00_init() )
	{
		// Read product code.
		value = s1d13c00_readReg( REG306C_PRODCODE );
		printf( "REG[%04Xh] Product Code = %04Xh\n", REG306C_PRODCODE&0xFFFF, value );
		if ( value != 0x005F )
		{
			printf( "***ERROR*** %04Xh is not the correct S1D13C00 product code (005Fh)!\n", value );
			goto EXIT_APP;
		}

		// Test scratchpad register.
		s1d13c00_writeReg( REG3060_MDCSCRATCHA0, 0x1234 );
		value = s1d13c00_readReg( REG3060_MDCSCRATCHA0 );
		printf( "REG[%04Xh] Scratchpad = %04Xh\n", REG3060_MDCSCRATCHA0&0xFFFF, value );
		if ( value != 0x1234 )
		{
			printf( "***ERROR*** Wrote %04Xh to REG[%04Xh], but read %04Xh afterwards!\n", value, REG3060_MDCSCRATCHA0&0xFFFF, value );
			goto EXIT_APP;
		}

		// Perform a memory write and read.
		int x = C00_DISPLAY_WIDTH / 2;
		int y = x;
		writePixel( x, y, MAGENTA );
		value = readPixel( x, y );
		printf( "Read pixel at (%d,%d) = %02Xh\n", x, y, value );
		if ( value != MAGENTA )
		{
			printf( "***ERROR*** Reading from memory did not work!\n" );
			goto EXIT_APP;
		}

		// Show three R G B bars in center of screen.
		printf( "Show RGB bars\n" );
		long t = getTime( );
		int barWidth  = 10;
		int barHeight = 40;
		x = (C00_DISPLAY_WIDTH-barHeight) / 2;
		y = (C00_DISPLAY_HEIGHT-barHeight) / 2;
		writeRect( x,    y, barWidth, barHeight, RED );
		writeRect( x+15, y, barWidth, barHeight, GREEN );
		writeRect( x+30, y, barWidth, barHeight, BLUE );
		printTimeDiff( t );

		// Draw a horizontal line above the bars, one pixel at a time via direct memory access.
		y -= 5;
		printf( "Draw a vertical line starting at (%d,%d)\n", x, y );
		for ( int i=0; i<barHeight; i++ )
			writePixel( x+i, y, MAGENTA );
		s1d13c00_updateDisplay( );

		// Read that line back and make sure the data is correct.
		u16* data = calloc( barHeight, sizeof(u16) );
		s1d13c00_readMemBurst( y*C00_DISPLAY_STRIDE+x, data, barHeight );
		for ( int i=0; i<barHeight; i++ )
		{
			if ( data[i] != MAGENTA )
			{
				printf( "***ERROR*** Writing to memory did not work!\n" );
				break;
			}
		}
		free( data );
	}

	EXIT_APP:
	s1d13c00_exit( );
	return 0;
}



